@@ -141,39 +141,6 @@ int disk_read_sectors(struct volume *volume, void *buf, uint64_t block, size_t c
     return DISK_SUCCESS;
 }
 
-static int disk_write_sectors(struct volume *volume, void *buf, uint64_t block, size_t count) {
-    struct dap dap = {0};
-
-    if (count * volume->sector_size > XFER_BUF_SIZE)
-        panic(false, "XFER");
-
-    if (xfer_buf == NULL)
-        xfer_buf = conv_mem_alloc(XFER_BUF_SIZE);
-
-    dap.size    = 16;
-    dap.count   = count;
-    dap.segment = rm_seg(xfer_buf);
-    dap.offset  = rm_off(xfer_buf);
-    dap.lba     = block;
-
-    struct rm_regs r = {0};
-    r.eax = 0x4301;
-    r.edx = volume->drive;
-    r.esi = (uint32_t)rm_off(&dap);
-    r.ds  = rm_seg(&dap);
-
-    if (buf != NULL)
-        memcpy(xfer_buf, buf, count * volume->sector_size);
-
-    rm_int(0x13, &r, &r);
-
-    if (r.eflags & EFLAGS_CF) {
-        return DISK_FAILURE;
-    }
-
-    return DISK_SUCCESS;
-}
-
 static bool detect_sector_size(struct volume *volume) {
     struct dap dap = {0};
 
@@ -230,12 +197,36 @@ static bool detect_sector_size(struct volume *volume) {
         }
     }
 
-    volume->sector_size = sector_size_a > sector_size_b ? sector_size_a : sector_size_b;
+    size_t detected_size = sector_size_a > sector_size_b ? sector_size_a : sector_size_b;
 
-    if (volume->sector_size == 0) {
+    if (detected_size == 0) {
         return false;
     }
 
+    // Validate detected sector size is a power of 2 and within reasonable bounds
+    // Valid sector sizes are typically 512, 1024, 2048, or 4096 bytes
+    static const size_t valid_sector_sizes[] = { 512, 1024, 2048, 4096 };
+    bool valid = false;
+    for (size_t i = 0; i < SIZEOF_ARRAY(valid_sector_sizes); i++) {
+        if (detected_size == valid_sector_sizes[i]) {
+            valid = true;
+            break;
+        }
+    }
+
+    if (!valid) {
+        // Round down to nearest valid sector size
+        volume->sector_size = 512;  // Default fallback
+        for (size_t i = SIZEOF_ARRAY(valid_sector_sizes); i > 0; i--) {
+            if (detected_size >= valid_sector_sizes[i - 1]) {
+                volume->sector_size = valid_sector_sizes[i - 1];
+                break;
+            }
+        }
+    } else {
+        volume->sector_size = detected_size;
+    }
+
     return true;
 }
 
@@ -280,18 +271,23 @@ void disk_create_index(void) {
         block->drive = drive;
         block->partition = 0;
         block->first_sect = 0;
-        block->sect_count = drive_params.lba_count;
         block->max_partition = -1;
 
         if (!detect_sector_size(block)) {
             continue;
         }
 
-        if (disk_read_sectors(block, xfer_buf, 0, 1) != DISK_SUCCESS) {
-            continue;
+        // Normalize sect_count to 512-byte sectors for consistency with partitions
+        // Preserve (uint64_t)-1 sentinel value (means "unknown size")
+        if (drive_params.lba_count == (uint64_t)-1 || drive_params.lba_count == 0) {
+            block->sect_count = (uint64_t)-1;
+        } else {
+            block->sect_count = drive_params.lba_count * (block->sector_size / 512);
         }
 
-        block->is_optical = (disk_write_sectors(block, xfer_buf, 0, 1) != DISK_SUCCESS) && block->sector_size == 2048;
+        // Detect optical drives via DPTE ATAPI bit (bit 6) or sector size heuristic
+        bool is_atapi = (dpte != NULL && (dpte->flags & (1 << 6)));
+        block->is_optical = is_atapi || (block->sector_size == 2048 && is_removable);
 
         if (!is_removable && !block->is_optical) {
             if (consumed_bda_disks == bda_disk_count) {
@@ -324,10 +320,14 @@ void disk_create_index(void) {
             struct volume *p = ext_mem_alloc(sizeof(struct volume));
             int ret = part_get(p, block, part);
 
-            if (ret == END_OF_TABLE || ret == INVALID_TABLE)
+            if (ret == END_OF_TABLE || ret == INVALID_TABLE) {
+                pmm_free(p, sizeof(struct volume));
                 break;
-            if (ret == NO_PARTITION)
+            }
+            if (ret == NO_PARTITION) {
+                pmm_free(p, sizeof(struct volume));
                 continue;
+            }
 
             volume_index = pmm_realloc(
                 volume_index,
@@ -393,6 +393,9 @@ static struct volume *pxe_from_efi_handle(EFI_HANDLE efi_handle) {
 
 #define UNIQUE_SECTOR_POOL_SIZE 65536
 static uint8_t *unique_sector_pool;
+static bool unique_sectors_calculated = false;
+
+static void find_unique_sectors(void);
 
 static struct volume *volume_by_unique_sector(void *b2b) {
     for (size_t i = 0; i < volume_index_i; i++) {
@@ -408,130 +411,269 @@ static struct volume *volume_by_unique_sector(void *b2b) {
     return NULL;
 }
 
-struct volume *disk_volume_from_efi_handle(EFI_HANDLE efi_handle) {
-    struct volume *ret;
+// Search for matching hash including invalidated volumes (for collision detection)
+static struct volume *volume_by_sector_hash(void *b2b) {
+    for (size_t i = 0; i < volume_index_i; i++) {
+        if (volume_index[i]->unique_sector_valid == false
+         && memcmp(volume_index[i]->unique_sector_b2b, (uint8_t[BLAKE2B_OUT_BYTES]){0}, BLAKE2B_OUT_BYTES) == 0) {
+            // Hash was never set, skip
+            continue;
+        }
 
+        if (memcmp(volume_index[i]->unique_sector_b2b, b2b, BLAKE2B_OUT_BYTES) == 0) {
+            return volume_index[i];
+        }
+    }
+
+    return NULL;
+}
+
+static bool is_efi_handle_to_skip(EFI_HANDLE efi_handle) {
+    EFI_STATUS status;
+
+    EFI_GUID dp_guid = EFI_DEVICE_PATH_PROTOCOL_GUID;
+    EFI_DEVICE_PATH_PROTOCOL *dp = NULL;
+
+    EFI_GUID guids_to_skip[] = {
+        // skip 7CCE9C94-983F-4D0A-8143-B6C05545B223 since it is apparently used by exposed
+        // ROM devices that we do not want to touch
+        // (see https://github.com/limine-bootloader/limine/issues/521#issuecomment-3160168795)
+        {0x7CCE9C94, 0x983F, 0x4D0A, {0x81, 0x43, 0xB6, 0xC0, 0x55, 0x45, 0xB2, 0x23}},
+    };
+
+    status = gBS->HandleProtocol(efi_handle, &dp_guid, (void **)&dp);
+    if (status) {
+        return false;
+    }
+
+    for (;;) {
+        if (dp->Type == END_DEVICE_PATH_TYPE && dp->SubType == END_ENTIRE_DEVICE_PATH_SUBTYPE) {
+            break;
+        }
+
+        uint16_t len = *(uint16_t *)dp->Length;
+
+        // Validate minimum device path node size before accessing type-specific data
+        if (len < sizeof(EFI_DEVICE_PATH_PROTOCOL)) {
+            break;  // Malformed device path node
+        }
+
+        if (dp->Type == HARDWARE_DEVICE_PATH && dp->SubType == HW_VENDOR_DP) {
+            // Vendor device path must be large enough to contain a GUID
+            if (len >= sizeof(EFI_DEVICE_PATH_PROTOCOL) + sizeof(EFI_GUID)) {
+                EFI_GUID *vendor_guid = (void *)dp + sizeof(EFI_DEVICE_PATH_PROTOCOL);
+
+                for (size_t i = 0; i < SIZEOF_ARRAY(guids_to_skip); i++) {
+                    if (memcmp(vendor_guid, &guids_to_skip[i], sizeof(EFI_GUID)) == 0) {
+                        return true;
+                    }
+                }
+            }
+        }
+        dp = (void *)dp + len;
+    }
+
+    return false;
+}
+
+static bool is_efi_handle_optical(EFI_HANDLE efi_handle) {
+    EFI_STATUS status;
+
+    EFI_GUID dp_guid = EFI_DEVICE_PATH_PROTOCOL_GUID;
+    EFI_DEVICE_PATH_PROTOCOL *dp = NULL;
+
+    status = gBS->HandleProtocol(efi_handle, &dp_guid, (void **)&dp);
+    if (status) {
+        return false;
+    }
+
+    for (;;) {
+        if (dp->Type == END_DEVICE_PATH_TYPE && dp->SubType == END_ENTIRE_DEVICE_PATH_SUBTYPE) {
+            break;
+        }
+
+        if (dp->Type == MEDIA_DEVICE_PATH && dp->SubType == MEDIA_CDROM_DP) {
+            return true;
+        }
+
+        uint16_t len = *(uint16_t *)dp->Length;
+        if (len < sizeof(EFI_DEVICE_PATH_PROTOCOL)) {
+            break;  // Malformed device path node
+        }
+        dp = (void *)dp + len;
+    }
+
+    return false;
+}
+
+static EFI_DEVICE_PATH_PROTOCOL *get_device_path(EFI_HANDLE efi_handle) {
+    EFI_STATUS status;
+    EFI_GUID dp_guid = EFI_DEVICE_PATH_PROTOCOL_GUID;
+    EFI_DEVICE_PATH_PROTOCOL *dp = NULL;
+
+    status = gBS->HandleProtocol(efi_handle, &dp_guid, (void **)&dp);
+    if (status) {
+        return NULL;
+    }
+    return dp;
+}
+
+// Compare device paths up to (but not including) partition nodes
+static bool device_paths_match_disk(EFI_DEVICE_PATH_PROTOCOL *dp1,
+                                    EFI_DEVICE_PATH_PROTOCOL *dp2) {
+    if (dp1 == NULL || dp2 == NULL) {
+        return false;
+    }
+
+    while (!IsDevicePathEnd(dp1) && !IsDevicePathEnd(dp2)) {
+        // Stop at partition nodes
+        if (dp1->Type == MEDIA_DEVICE_PATH &&
+            (dp1->SubType == MEDIA_HARDDRIVE_DP || dp1->SubType == MEDIA_CDROM_DP)) {
+            break;
+        }
+        if (dp2->Type == MEDIA_DEVICE_PATH &&
+            (dp2->SubType == MEDIA_HARDDRIVE_DP || dp2->SubType == MEDIA_CDROM_DP)) {
+            break;
+        }
+
+        uint16_t len1 = DevicePathNodeLength(dp1);
+        uint16_t len2 = DevicePathNodeLength(dp2);
+
+        if (len1 != len2) {
+            return false;
+        }
+
+        if (len1 < sizeof(EFI_DEVICE_PATH_PROTOCOL)) {
+            return false;
+        }
+
+        if (memcmp(dp1, dp2, len1) != 0) {
+            return false;
+        }
+
+        dp1 = (void *)dp1 + len1;
+        dp2 = (void *)dp2 + len2;
+    }
+
+    return true;
+}
+
+static struct volume *volume_by_device_path(EFI_HANDLE query_handle) {
+    EFI_DEVICE_PATH_PROTOCOL *query_dp = get_device_path(query_handle);
+    if (query_dp == NULL) {
+        return NULL;
+    }
+
+    for (size_t i = 0; i < volume_index_i; i++) {
+        EFI_DEVICE_PATH_PROTOCOL *vol_dp = get_device_path(volume_index[i]->efi_handle);
+        if (vol_dp == NULL) {
+            continue;
+        }
+
+        if (device_paths_match_disk(query_dp, vol_dp)) {
+            // Convert first_sect from 512-byte sectors to device LBAs
+            int sector_size = volume_index[i]->sector_size;
+            if ((volume_index[i]->first_sect * 512) % sector_size) {
+                continue;  // Misaligned, skip this volume
+            }
+            uint64_t first_sect_lba = (volume_index[i]->first_sect * 512) / sector_size;
+
+            EFI_DEVICE_PATH_PROTOCOL *qp = query_dp;
+            while (!IsDevicePathEnd(qp)) {
+                if (qp->Type == MEDIA_DEVICE_PATH && qp->SubType == MEDIA_HARDDRIVE_DP) {
+                    uint16_t len = DevicePathNodeLength(qp);
+                    // UEFI spec size is 42 bytes, but sizeof() may be larger due to padding
+                    if (len < 42) {
+                        break;
+                    }
+                    HARDDRIVE_DEVICE_PATH *query_hd = (HARDDRIVE_DEVICE_PATH *)qp;
+                    if (first_sect_lba == query_hd->PartitionStart) {
+                        return volume_index[i];
+                    }
+                    break;
+                }
+                if (qp->Type == MEDIA_DEVICE_PATH && qp->SubType == MEDIA_CDROM_DP) {
+                    uint16_t len = DevicePathNodeLength(qp);
+                    if (len < sizeof(CDROM_DEVICE_PATH)) {
+                        break;
+                    }
+                    CDROM_DEVICE_PATH *query_cd = (CDROM_DEVICE_PATH *)qp;
+                    if (first_sect_lba == query_cd->PartitionStart) {
+                        return volume_index[i];
+                    }
+                    break;
+                }
+                uint16_t len = DevicePathNodeLength(qp);
+                if (len < sizeof(EFI_DEVICE_PATH_PROTOCOL)) {
+                    break;
+                }
+                qp = (void *)qp + len;
+            }
+
+            if (IsDevicePathEnd(qp) && volume_index[i]->partition == 0) {
+                return volume_index[i];
+            }
+        }
+    }
+
+    return NULL;
+}
+
+struct volume *disk_volume_from_efi_handle(EFI_HANDLE efi_handle) {
     EFI_STATUS status;
 
     EFI_GUID block_io_guid = BLOCK_IO_PROTOCOL;
     EFI_BLOCK_IO *block_io = NULL;
 
+    if (is_efi_handle_to_skip(efi_handle)) {
+        return NULL;
+    }
+
     status = gBS->HandleProtocol(efi_handle, &block_io_guid, (void **)&block_io);
     if (status) {
         return pxe_from_efi_handle(efi_handle);
     }
 
-    block_io->Media->WriteCaching = false;
+    // Try device path matching first (primary method)
+    struct volume *ret = volume_by_device_path(efi_handle);
+    if (ret != NULL) {
+        return ret;
+    }
 
+    // Fallback to unique sector matching
     uint64_t bdev_size = ((uint64_t)block_io->Media->LastBlock + 1) * (uint64_t)block_io->Media->BlockSize;
-    if (bdev_size < UNIQUE_SECTOR_POOL_SIZE) {
-        goto fallback;
-    }
+    if (bdev_size >= UNIQUE_SECTOR_POOL_SIZE) {
+        status = block_io->ReadBlocks(block_io, block_io->Media->MediaId,
+                                      0,
+                                      UNIQUE_SECTOR_POOL_SIZE,
+                                      unique_sector_pool);
+        if (status == 0) {
+            find_unique_sectors();
 
-    status = block_io->ReadBlocks(block_io, block_io->Media->MediaId,
-                                  0,
-                                  UNIQUE_SECTOR_POOL_SIZE,
-                                  unique_sector_pool);
-    if (status != 0) {
-        goto fallback;
-    }
+            uint8_t b2b[BLAKE2B_OUT_BYTES];
+            blake2b(b2b, unique_sector_pool, UNIQUE_SECTOR_POOL_SIZE);
 
-    uint8_t b2b[BLAKE2B_OUT_BYTES];
-    blake2b(b2b, unique_sector_pool, UNIQUE_SECTOR_POOL_SIZE);
-
-    ret = volume_by_unique_sector(b2b);
-    if (ret != NULL) {
-        return ret;
-    }
-
-    // Fallback to read-back method
-fallback:;
-    uint64_t signature = rand64();
-    uint64_t new_signature;
-    do { new_signature = rand64(); } while (new_signature == signature);
-    uint64_t orig;
-
-    status = block_io->ReadBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-    orig = *(uint64_t *)unique_sector_pool;
-    if (status) {
-        return NULL;
-    }
-
-    *(uint64_t *)unique_sector_pool = signature;
-    status = block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-    if (status) {
-        return NULL;
-    }
-
-    ret = NULL;
-    for (size_t i = 0; i < volume_index_i; i++) {
-        uint64_t compare;
-
-        status = volume_index[i]->block_io->ReadBlocks(volume_index[i]->block_io,
-                          volume_index[i]->block_io->Media->MediaId,
-                          (volume_index[i]->first_sect * 512) / volume_index[i]->sector_size,
-                          4096, unique_sector_pool);
-        compare = *(uint64_t *)unique_sector_pool;
-        if (status) {
-            continue;
-        }
-
-        if (compare == signature) {
-            // Double check
-            status = block_io->ReadBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-            if (status) {
-                break;
-            }
-            *(uint64_t *)unique_sector_pool = new_signature;
-            status = block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-            if (status) {
-                break;
-            }
-
-            status = volume_index[i]->block_io->ReadBlocks(volume_index[i]->block_io,
-                          volume_index[i]->block_io->Media->MediaId,
-                          (volume_index[i]->first_sect * 512) / volume_index[i]->sector_size,
-                          4096, unique_sector_pool);
-            compare = *(uint64_t *)unique_sector_pool;
-            if (status) {
-                continue;
-            }
-
-            if (compare == new_signature) {
-                ret = volume_index[i];
-                break;
-            }
-
-            status = block_io->ReadBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-            if (status) {
-                break;
-            }
-            *(uint64_t *)unique_sector_pool = signature;
-            status = block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-            if (status) {
-                break;
+            ret = volume_by_unique_sector(b2b);
+            if (ret != NULL) {
+                // Verify size, block size, and partition status match
+                if (block_io->Media->BlockSize == (uint32_t)ret->sector_size
+                 && bdev_size == ret->sect_count * 512
+                 && block_io->Media->LogicalPartition == (ret->partition != 0)) {
+                    return ret;
+                }
             }
         }
     }
 
-    status = block_io->ReadBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-    if (status) {
-        return NULL;
-    }
-    *(uint64_t *)unique_sector_pool = orig;
-    status = block_io->WriteBlocks(block_io, block_io->Media->MediaId, 0, 4096, unique_sector_pool);
-    if (status) {
-        return NULL;
-    }
-
-    if (ret != NULL) {
-        return ret;
-    }
-
     return NULL;
 }
 
 static void find_unique_sectors(void) {
+    if (unique_sectors_calculated) {
+        return;
+    }
+    unique_sectors_calculated = true;
+
     EFI_STATUS status;
 
     for (size_t i = 0; i < volume_index_i; i++) {
@@ -541,7 +683,8 @@ static void find_unique_sectors(void) {
 
         size_t first_sect = (volume_index[i]->first_sect * 512) / volume_index[i]->sector_size;
 
-        if (volume_index[i]->sect_count * volume_index[i]->sector_size < UNIQUE_SECTOR_POOL_SIZE) {
+        // sect_count is always in 512-byte sectors
+        if (volume_index[i]->sect_count * 512 < UNIQUE_SECTOR_POOL_SIZE) {
             continue;
         }
 
@@ -558,15 +701,21 @@ static void find_unique_sectors(void) {
         uint8_t b2b[BLAKE2B_OUT_BYTES];
         blake2b(b2b, unique_sector_pool, UNIQUE_SECTOR_POOL_SIZE);
 
-        struct volume *collision = volume_by_unique_sector(b2b);
+        // Check for collision BEFORE storing hash (so we don't find ourselves)
+        // This searches all volumes including previously invalidated ones
+        struct volume *collision = volume_by_sector_hash(b2b);
+
+        // Always store the hash so future volumes can detect collisions
+        memcpy(volume_index[i]->unique_sector_b2b, b2b, BLAKE2B_OUT_BYTES);
+
         if (collision == NULL) {
             volume_index[i]->unique_sector_valid = true;
-            memcpy(volume_index[i]->unique_sector_b2b, b2b, BLAKE2B_OUT_BYTES);
             continue;
         }
 
-        // Invalidate collision's unique sector
+        // Collision found - invalidate both volumes
         collision->unique_sector_valid = false;
+        volume_index[i]->unique_sector_valid = false;
     }
 }
 
@@ -621,6 +770,10 @@ fail:
     for (size_t i = 0; i < handle_count; i++) {
         EFI_BLOCK_IO *drive = NULL;
 
+        if (is_efi_handle_to_skip(handles[i])) {
+            continue;
+        }
+
         status = gBS->HandleProtocol(handles[i], &block_io_guid, (void **)&drive);
 
         if (status != 0 || drive == NULL || drive->Media->LastBlock == 0)
@@ -629,18 +782,18 @@ fail:
         if (drive->Media->LogicalPartition)
             continue;
 
-        drive->Media->WriteCaching = false;
-
+        // Read test to ensure device is responsive (skipping this causes hangs on some systems)
         status = drive->ReadBlocks(drive, drive->Media->MediaId, 0, 4096, unique_sector_pool);
         if (status) {
             continue;
         }
 
-        status = drive->WriteBlocks(drive, drive->Media->MediaId, 0, 4096, unique_sector_pool);
-
         struct volume *block = ext_mem_alloc(sizeof(struct volume));
 
-        if ((status || drive->Media->ReadOnly) && drive->Media->BlockSize == 2048) {
+        bool is_optical = is_efi_handle_optical(handles[i]) ||
+                          (drive->Media->ReadOnly && drive->Media->BlockSize == 2048);
+
+        if (is_optical) {
             block->index = optical_indices++;
             block->is_optical = true;
         } else {
@@ -652,7 +805,8 @@ fail:
         block->partition = 0;
         block->sector_size = drive->Media->BlockSize;
         block->first_sect = 0;
-        block->sect_count = drive->Media->LastBlock + 1;
+        // Normalize sect_count to 512-byte sectors for consistency with partitions
+        block->sect_count = (drive->Media->LastBlock + 1) * (drive->Media->BlockSize / 512);
         block->max_partition = -1;
 
         if (drive->Revision >= EFI_BLOCK_IO_PROTOCOL_REVISION3) {
@@ -700,7 +854,6 @@ fail:
         }
     }
 
-    find_unique_sectors();
     find_part_handles(handles, handle_count);
 
     pmm_free(handles, handles_size);
