@@ -385,10 +385,14 @@ __attribute__((always_inline)) static inline void genloop(struct fb_info *fb, si
             }
             else { /* internal part */
                 for (size_t x = xstart; x < xend; x++) {
-                    size_t image_x = (x - background->x_displacement);
-                    bool x_external = image_x >= background->x_size;
-                    uint32_t img_pixel = *(uint32_t*)(img + image_x * colsize + off);
-                    uint32_t i = blend(fb, x, y, x_external ? background->back_colour : img_pixel);
+                    uint32_t pixel;
+                    if (x < background->x_displacement || x - background->x_displacement >= background->x_size) {
+                        pixel = background->back_colour;
+                    } else {
+                        size_t image_x = x - background->x_displacement;
+                        pixel = *(uint32_t*)(img + image_x * colsize + off);
+                    }
+                    uint32_t i = blend(fb, x, y, pixel);
                     bg_canvas[canvas_off + x] = i;
                 }
             }
@@ -426,6 +430,10 @@ static void loop_internal(struct fb_info *fb, size_t xstart, size_t xend, size_t
 
 static void generate_canvas(struct fb_info *fb) {
     if (background) {
+        // Free previous canvas if it exists
+        if (bg_canvas != NULL) {
+            pmm_free(bg_canvas, bg_canvas_size);
+        }
         bg_canvas_size = fb->framebuffer_width * fb->framebuffer_height * sizeof(uint32_t);
         bg_canvas = ext_mem_alloc(bg_canvas_size);
 
@@ -469,6 +477,13 @@ static void riscv_flush_callback(volatile void *base, size_t length) {
         asm volatile("cbo.flush (%0)" :: "r"(ptr) : "memory");
     }
 }
+static void riscv_flush_callback_nozicbom(volatile void *base, size_t length) {
+    volatile uint64_t *p = (volatile uint64_t *)base;
+    for (size_t i = 0; i < (length * 2) / sizeof(uint64_t); i += (64 / sizeof(uint64_t))) {
+        (void)p[i];
+    }
+    asm volatile ("fence r, r" ::: "memory");
+}
 #elif defined (__aarch64__)
 static void aarch64_flush_callback(volatile void *base, size_t length) {
     clean_dcache_poc((uintptr_t)base, (uintptr_t)base + length);
@@ -839,6 +854,8 @@ no_load_font:;
 #if defined (__riscv)
         if (riscv_check_isa_extension("zicbom", NULL, NULL)) {
             flanterm_fb_set_flush_callback(term, riscv_flush_callback);
+        } else {
+            flanterm_fb_set_flush_callback(term, riscv_flush_callback_nozicbom);
         }
 #elif defined (__aarch64__)
         flanterm_fb_set_flush_callback(term, aarch64_flush_callback);
